
import string
import yaml


def _validate_mailing_method(value):
    methods = ['stdout', 'smtp']
    if value not in methods:
        return f'permitted values: {methods}, got {value}'


def _validate_smtp_config(value):
    if not isinstance(value, dict):
        return f'must be a map, got {type(value)}'
    if not isinstance(value['host'], str):
        return f'host must be a str, got {type(value["host"])}'
    if not isinstance(value['port'], int):
        return f'port must be a int, got {type(value["port"])}'
    if 'tls' in value:
        if not isinstance(value['tls'], bool):
            return f'tls must be a bool, got {type(value["tls"])}'
    else:
        value['tls'] = False
    if 'starttls' in value:
        if not isinstance(value['starttls'], bool):
            return f'starttls must be a bool, got {type(value["starttls"])}'
    else:
        value['starttls'] = False
    if 'username' in value and 'password' in value:
        if not isinstance(value['username'], str):
            return f'username must be a str, got {type(value["username"])}'
        if not isinstance(value['password'], str):
            return f'password must be a str, got {type(value["password"])}'
    else:
        value['username'] = None
        value['password'] = None


def _validate_httpd_config(value):
    if not isinstance(value, dict):
        return f'must be a map, got {type(value)}'
    if not isinstance(value['host'], str):
        return f'host must be a str, got {type(value["host"])}'
    if not isinstance(value['port'], int):
        return f'port must be a int, got {type(value["port"])}'


def _validate_lmtpd_config(value):
    if not isinstance(value, dict):
        return f'must be a map, got {type(value)}'
    if not isinstance(value['host'], str):
        return f'host must be a str, got {type(value["host"])}'
    if not isinstance(value['port'], int):
        return f'port must be a int, got {type(value["port"])}'


def _validate_policy_flags(value):
    alphabet = string.ascii_lowercase + string.digits + '-._'
    if not isinstance(value, dict):
        return f'must be a map, got {type(value)}'
    for flag, v in value.items():
        if not isinstance(flag, str) or len(flag) == 0:
            return 'has non-string or empty members'
        if flag[0] not in string.ascii_lowercase:
            return 'must start with a lowercase letter'
        for c in flag:
            if c not in alphabet:
                return f'has invalid key {flag}'


class _ConfigOption:

    def __init__(self, key, typ, default, validator=None):
        self._key = key
        self._typ = typ
        self._default = default
        self._validator = validator
        self._value = default

    @property
    def value(self):
        return self._value

    def load(self, conf: dict):
        self._value = conf.get(self._key, self._default)
        if not isinstance(self._value, self._typ):
            raise TypeError(f'Option {self._key} must be {self._typ}, got {type(self._value)}')
        if self._validator:
            response = self._validator(self._value)
            if response:
                raise ValueError(f'Invalid value for option {self._key}: {response}')


class _Config:

    def __init__(self, **kwargs):
        self._options = kwargs

    def __getattr__(self, item):
        if item in self._options:
            return self._options[item].value
        raise AttributeError()


class _GlobalConfig(_Config):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__domains = {}

    def __make_domain(self, domain):
        self.__domains[domain] = _Config(
            submission_address=_ConfigOption('address', str, f'gpgwks@{domain}'),
            passphrase=_ConfigOption('passphrase', str, ''),
            policy_flags=_ConfigOption('policy_flags', dict, {}, validator=_validate_policy_flags)
        )

    def __getitem__(self, item):
        return self.__domains.get(item)

    @property
    def domains(self) -> list:
        return list(self.__domains.keys())

    def load_config(self, path: str = '/etc/easywks.yml'):
        with open(path, 'r') as f:
            conf = yaml.safe_load(f)
        for co in self._options.values():
            co.load(conf)
        self.__domains = {}
        for domain, dconf in conf['domains'].items():
            self.__make_domain(domain)
            for co in self.__domains[domain]._options.values():
                co.load(conf)


Config = _GlobalConfig(
    working_directory=_ConfigOption('directory', str, '/var/lib/easywks'),
    pending_lifetime=_ConfigOption('pending_lifetime', int, 604800),
    mailing_method=_ConfigOption('mailing_method', str, 'stdout', validator=_validate_mailing_method),
    permit_unsigned_response=_ConfigOption('permit_unsigned_response', bool, False),
    httpd=_ConfigOption('httpd', dict, {
        'host': 'localhost',
        'port': 8080
    }, validator=_validate_httpd_config),
    smtp=_ConfigOption('smtp', dict, {
        'host': 'localhost',
        'port': 25,
        'tls': False,
        'starttls': False
    }, validator=_validate_smtp_config),
    lmtpd=_ConfigOption('lmtpd', dict, {
        'host': 'localhost',
        'port': 25,
    }, validator=_validate_lmtpd_config),
)
