
from datetime import datetime
from email.policy import default
from email.utils import format_datetime
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText

from pgpy import PGPKey, PGPMessage, PGPUID
from pgpy.types import SignatureVerification

from .crypto import pgp_sign
from .config import Config
from .util import create_nonce, fingerprint


XLOOP_HEADER = 'EasyWKS'


class SubmissionRequest:

    def __init__(self, submitter_addr: str, submission_addr: str, key: PGPKey):
        self._submitter_addr = submitter_addr
        self._submission_addr = submission_addr
        self._key = key

    def confirmation_request(self) -> 'ConfirmationRequest':
        return ConfirmationRequest(self._submitter_addr, self. _submission_addr, self._key)

    @property
    def submitter_address(self):
        return self._submitter_addr

    @property
    def submission_address(self):
        return self._submission_addr

    @property
    def key(self):
        return self._key


class ConfirmationRequest:

    MAIL_TEXT = '''Hi there!

This is the EasyWKS system at {domain}.

You appear to have submitted your key for publication in the Web Key
Directory.  There's one more step you need to complete.  If you did not
request this, you can simply ignore this message.

If your email client doesn't automatically complete this challenge, you
can perform this step manually: Please verify that you can decrypt the
second part of this message and that the fingerprint listed in the
encrypted part matches your key.  If everything looks ok, please reply
to this message with an **encrypted and signed PGP/MIME message** with
the following content (without the <> brackets)

  type: confirmation-response
  sender: <your email address>
  nonce: <copy the nonce from the encrypted part of this message>

For more information on WKD and WKS see:

  https://gnupg.org/faq/wkd.html
  https://gnupg.org/faq/wks.html


Regards
EasyWKS

--
Dance like nobody is watching.
Encrypt live everybody is.
'''

    def __init__(self, submitter_addr: str, submission_addr: str, key: PGPKey, nonce: str = None):
        self._domain = submitter_addr.split('@')[1]
        self._submitter_addr = submitter_addr
        self._submission_addr = submission_addr
        self._key = key
        self._nonce = nonce or create_nonce()

    @property
    def domain(self):
        return self._domain

    @property
    def submitter_address(self):
        return self._submitter_addr

    @property
    def submission_address(self):
        return self._submission_addr

    @property
    def key(self):
        return self._key

    @property
    def nonce(self):
        return self._nonce

    def create_signed_message(self):
        mpplain = MIMEText(ConfirmationRequest.MAIL_TEXT.format(domain=self.domain), _subtype='plain')
        ps = '\r\n'.join([
            'type: confirmation-request',
            f'sender: {self._submission_addr}',
            f'address: {self._submitter_addr}',
            f'fingerprint: {fingerprint(self._key)}',
            f'nonce: {self._nonce}',
            ''
        ])
        payload = MIMEText(ps, _subtype='plain')
        to_encrypt = PGPMessage.new(payload.as_string(policy=default))
        encrypted = self._key.encrypt(to_encrypt)
        mpenc = MIMEApplication(str(encrypted), _subtype='vnd.gnupg.wks')
        mixed = MIMEMultipart(_subtype='mixed', _subparts=[mpplain, mpenc])
        to_sign = PGPMessage.new(mixed.as_string(policy=default))
        sig = pgp_sign(self.domain, to_sign)
        mpsig = MIMEApplication(str(sig), _subtype='pgp-signature')
        email = MIMEMultipart(_subtype='signed', _subparts=[mixed, mpsig], policy=default,
                              protocol='application/pgp-signature')
        email['Subject'] = 'Confirm your key publication'
        email['To'] = self._submitter_addr
        email['From'] = self._submission_addr
        email['Date'] = format_datetime(datetime.utcnow())
        email['Wks-Draft-Version'] = '3'
        email['Wks-Phase'] = 'confirm'
        email['X-Loop'] = XLOOP_HEADER
        email['Auto-Submitted'] = 'auto-replied'
        return email


class ConfirmationResponse:

    def __init__(self, submitter_addr: str, submission_addr: str, nonce: str, msg: PGPMessage):
        self._domain = submitter_addr.split('@')[1]
        self._submitter_addr = submitter_addr
        self._submission_addr = submission_addr
        self._nonce = nonce
        self._msg = msg

    @property
    def domain(self):
        return self._domain

    @property
    def submitter_address(self):
        return self._submitter_addr

    @property
    def submission_address(self):
        return self._submission_addr

    @property
    def nonce(self):
        return self._nonce

    def get_publish_response(self, key: PGPKey) -> 'PublishResponse':
        return PublishResponse(self._submitter_addr, self._submission_addr, key)

    def verify_signature(self, key: PGPKey):
        if not self._msg.is_signed:
            if not Config.permit_unsigned_response:
                raise EasyWksError('The confirmation response is not signed. If you used an automated tool such as '
                                   'gpg-wks-client for submitting your response, please update said tool or try '
                                   'responding manually.')
            else:
                # Unsigned, but permitted
                return
        uid: PGPUID = key.get_uid(self._submitter_addr)
        if uid is None or uid.email != self._submitter_addr:
            raise EasyWksError(f'UID {self._submitter_addr} not found in PGP key')
        verification: SignatureVerification = key.verify(self._msg)
        for verified, by, sig, subject in verification.good_signatures:
            if fingerprint(key) == fingerprint(by):
                return
        raise EasyWksError('PGP signature could not be verified')


class PublishResponse:
    MAIL_TEXT = '''Hi there!

This is the EasyWKS system at {domain}.

Your key has been published to the Web Key Directory.
You can test WKD key retrieval e.g. with:

gpg --auto-key-locate=wkd,nodefault --locate-key {uid}

For more information on WKD and WKS see:

  https://gnupg.org/faq/wkd.html
  https://gnupg.org/faq/wks.html

Regards
EasyWKS

--
Dance like nobody is watching.
Encrypt live everybody is.
'''

    def __init__(self, submitter_addr: str, submission_addr: str, key: PGPKey):
        self._domain = submitter_addr.split('@')[1]
        self._submitter_addr = submitter_addr
        self._submission_addr = submission_addr
        self._key = key

    @property
    def submitter_address(self):
        return self._submitter_addr

    @property
    def submission_address(self):
        return self._submission_addr

    @property
    def key(self):
        return self._key

    @property
    def domain(self):
        return self._domain

    def create_signed_message(self):
        mpplain = MIMEText(PublishResponse.MAIL_TEXT.format(domain=self.domain, uid=self.submitter_address),
                           _subtype='plain')
        to_encrypt = PGPMessage.new(mpplain.as_string(policy=default))
        encrypted: PGPMessage = self.key.encrypt(to_encrypt)
        encrypted |= pgp_sign(self.domain, encrypted)
        payload = MIMEApplication(str(encrypted), _subtype='octet-stream')
        mpenc = MIMEApplication('Version: 1\r\n', _subtype='pgp-encrypted')
        email = MIMEMultipart(_subtype='encrypted', _subparts=[mpenc, payload], policy=default,
                              protocol='application/pgp-encrypted')
        email['Subject'] = 'Your key has been published'
        email['To'] = self.submitter_address
        email['From'] = self.submission_address
        email['Date'] = format_datetime(datetime.utcnow())
        email['Wks-Draft-Version'] = '3'
        email['Wks-Phase'] = 'done'
        email['X-Loop'] = XLOOP_HEADER
        email['Auto-Submitted'] = 'auto-replied'
        return email


class EasyWksError(BaseException):

    MAIL_TEXT = '''Hi there!

This is the EasyWKS system at {domain}.

An error has occurred while processing your request.

{message}

If this error persists, please contact your administrator for help.

For more information on WKD and WKS see:

  https://gnupg.org/faq/wkd.html
  https://gnupg.org/faq/wks.html


Regards
EasyWKS

--
Dance like nobody is watching.
Encrypt live everybody is.
'''

    def __init__(self, msg: str, ):
        super().__init__()
        self._msg = msg

    def __str__(self) -> str:
        return self._msg

    def create_message(self, submitter_addr: str, submission_addr: str) -> MIMEText:
        domain = submission_addr.split('@', 1)[1]
        payload = EasyWksError.MAIL_TEXT.format(domain=domain, message=self._msg)
        email = MIMEText(payload)
        email['Subject'] = 'An error has occurred while processing your request'
        email['From'] = submission_addr
        email['To'] = submitter_addr
        email['Date'] = format_datetime(datetime.utcnow())
        email['Wks-Draft-Version'] = '3'
        email['Wks-Phase'] = 'error'
        email['X-Loop'] = XLOOP_HEADER
        email['Auto-Submitted'] = 'auto-replied'
        return email
